/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997-2006 Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Sam Lantinga
    slouken@libsdl.org
*/
#include "SDL_config.h"

/* Dummy SDL video driver implementation; this is just enough to make an
 *  SDL-based application THINK it's got a working video driver, for
 *  applications that call SDL_Init(SDL_INIT_VIDEO) when they don't need it,
 *  and also for use as a collection of stubs when porting SDL to a new
 *  platform for which you haven't yet written a valid video driver.
 *
 * This is also a great way to determine bottlenecks: if you think that SDL
 *  is a performance problem for a given platform, enable this driver, and
 *  then see if your application runs faster without video overhead.
 *
 * Initial work by Ryan C. Gordon (icculus@icculus.org). A good portion
 *  of this was cut-and-pasted from Stephane Peter's work in the AAlib
 *  SDL video driver.  Renamed to "DUMMY" by Sam Lantinga.
 */

#include "SDL_video.h"
#include "SDL_mouse.h"
#include "../SDL_sysvideo.h"
#include "../SDL_pixels_c.h"
#include "../../events/SDL_events_c.h"

#include "SDL_xtermvideo.h"
#include "SDL_xtermevents_c.h"
#include "SDL_xtermmouse_c.h"

#include <ncurses.h>

#define XTERMVID_DRIVER_NAME "xterm"

/* Initialization/Query functions */
static int XTERM_VideoInit(_THIS, SDL_PixelFormat *vformat);
static SDL_Rect **XTERM_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags);
static SDL_Surface *XTERM_SetVideoMode(_THIS, SDL_Surface *current, int width, int height, int bpp, Uint32 flags);
static int XTERM_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors);
static void XTERM_VideoQuit(_THIS);

/* Hardware surface functions */
static int XTERM_AllocHWSurface(_THIS, SDL_Surface *surface);
static int XTERM_LockHWSurface(_THIS, SDL_Surface *surface);
static void XTERM_UnlockHWSurface(_THIS, SDL_Surface *surface);
static void XTERM_FreeHWSurface(_THIS, SDL_Surface *surface);

/* etc. */
static void XTERM_UpdateRects(_THIS, int numrects, SDL_Rect *rects);

/*
static int getenv_int(const char* name) {
    char* env = SDL_getenv(name);
    if (!env) {
        SDL_SetError("environment variable %s is not set.");
        return -1;
    }
    return atoi(env);
}
*/

/* XTERM driver bootstrap functions */

static int XTERM_Available(void)
{
	const char *envr = SDL_getenv("SDL_VIDEODRIVER");
	if ((envr) && (SDL_strcmp(envr, XTERMVID_DRIVER_NAME) == 0)) {
        return(1);
	}
	return(0);
}

static void XTERM_DeleteDevice(SDL_VideoDevice *device)
{
	SDL_free(device->hidden);
	SDL_free(device);
}

static SDL_VideoDevice *XTERM_CreateDevice(int devindex)
{
	SDL_VideoDevice *device;

	/* Initialize all variables that we clean on shutdown */
	device = (SDL_VideoDevice *)SDL_malloc(sizeof(SDL_VideoDevice));
	if ( device ) {
		SDL_memset(device, 0, (sizeof *device));
		device->hidden = (struct SDL_PrivateVideoData *)
				SDL_malloc((sizeof *device->hidden));
	}
	if ( (device == NULL) || (device->hidden == NULL) ) {
		SDL_OutOfMemory();
		if ( device ) {
			SDL_free(device);
		}
		return(0);
	}
	SDL_memset(device->hidden, 0, (sizeof *device->hidden));

	/* Set the function pointers */
	device->VideoInit = XTERM_VideoInit;
	device->ListModes = XTERM_ListModes;
	device->SetVideoMode = XTERM_SetVideoMode;
	device->CreateYUVOverlay = NULL;
	device->SetColors = XTERM_SetColors;
	device->UpdateRects = XTERM_UpdateRects;
	device->VideoQuit = XTERM_VideoQuit;
	device->AllocHWSurface = XTERM_AllocHWSurface;
	device->CheckHWBlit = NULL;
	device->FillHWRect = NULL;
	device->SetHWColorKey = NULL;
	device->SetHWAlpha = NULL;
	device->LockHWSurface = XTERM_LockHWSurface;
	device->UnlockHWSurface = XTERM_UnlockHWSurface;
	device->FlipHWSurface = NULL;
	device->FreeHWSurface = XTERM_FreeHWSurface;
	device->SetCaption = NULL;
	device->SetIcon = NULL;
	device->IconifyWindow = NULL;
	device->GrabInput = NULL;
	device->GetWMInfo = NULL;
	device->InitOSKeymap = XTERM_InitOSKeymap;
	device->PumpEvents = XTERM_PumpEvents;

	device->free = XTERM_DeleteDevice;

	return device;
}

VideoBootStrap XTERM_bootstrap = {
	XTERMVID_DRIVER_NAME, "SDL xterm video driver",
	XTERM_Available, XTERM_CreateDevice
};


int XTERM_VideoInit(_THIS, SDL_PixelFormat *vformat)
{
    /*
    int bpp = getenv_int("SDLXTERM_BPP");
    */

	vformat->BitsPerPixel = 0;
	vformat->BytesPerPixel = 0;

	/* We're done! */
	return(0);
}

SDL_Rect **XTERM_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags)
{
    return (SDL_Rect **) -1;
}

SDL_Surface *XTERM_SetVideoMode(_THIS, SDL_Surface *current,
                                int width, int height, int bpp, Uint32 flags)
{
    int r, g, b;
    WINDOW* win = initscr();
    this->hidden->win = win;
    this->hidden->cur_w = getmaxx(win);
    this->hidden->cur_h = getmaxy(win);
    this->hidden->w = width;
    this->hidden->h = height;
    this->hidden->bpp = bpp;

    nodelay(win, 1);
    keypad(win, 1);

    if ( this->hidden->buffer ) {
        SDL_free( this->hidden->buffer );
    }

    this->hidden->buffer = SDL_malloc(width * height * (bpp / 8));
    if ( ! this->hidden->buffer ) {
        SDL_SetError("Couldn't allocate buffer for requested mode");
        return(NULL);
    }

    SDL_memset(this->hidden->buffer, 0, width * height * (bpp / 8));

    /* Allocate the new pixel format for the screen */
    if ( ! SDL_ReallocFormat(current, bpp, 0, 0, 0, 0) ) {
        SDL_free(this->hidden->buffer);
        this->hidden->buffer = NULL;
        SDL_SetError("Couldn't allocate new pixel format for requested mode");
        return(NULL);
    }
    this->hidden->fmt = current->format;

	/* Set up the new mode framebuffer */
	//current->flags = SDL_SWSURFACE;
    current->flags = SDL_SWSURFACE | SDL_HWPALETTE;
    this->hidden->set_colored = 0;
    this->hidden->text = getenv("SDL_XTERM_TEXT");
    if (this->hidden->text == NULL) {
        this->hidden->text_len = 0;
    }
    else {
        this->hidden->text_len = strlen(this->hidden->text);
    }
	current->w = width;
	current->h = height;
    this->hidden->pitch = current->pitch = current->w * (bpp / 8);
	current->pixels = this->hidden->buffer;

#if 0   /* i cant */
    /* clipping? */
    if (x < 0 || y < 0 || x+width > w || y+height > h) {
        SDL_Rect c;
        c.x = (x < 0) ? -x : 0;
        c.w = (width > w-x) ? w-x : width;
        c.y = (y < 0) ? -y : 0;
        c.h = (height > h-y) ? h-y : height;
        current->clip_rect = c;
//        SDL_SetClipRect(current, &c);
        printf("%d %d %d %d\n", c.x, c.y, c.w, c.h);
    }
#endif

	/* We're done */
	return(current);
}

/* We don't actually allow hardware surfaces other than the main one */
static int XTERM_AllocHWSurface(_THIS, SDL_Surface *surface)
{
	return(-1);
}
static void XTERM_FreeHWSurface(_THIS, SDL_Surface *surface)
{
	return;
}

/* We need to wait for vertical retrace on page flipped displays */
static int XTERM_LockHWSurface(_THIS, SDL_Surface *surface)
{
	return(0);
}

static void XTERM_UnlockHWSurface(_THIS, SDL_Surface *surface)
{
	return;
}

static Uint32 getpixel(void* buf, int bpp, int pitch, int x, int y)
{
    Uint8 *p = (Uint8 *)buf + y * pitch + x * bpp;

    switch(bpp) {
    case 1:
        return *p;

    case 2:
        return *(Uint16 *)p;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
            return p[0] << 16 | p[1] << 8 | p[2];
        else
            return p[0] | p[1] << 8 | p[2] << 16;

    case 4:
        return *(Uint32 *)p;

    default:
        return 0;
    }
}

static int conv1(int v) {
    return v < 48 ? 0 : v < 115 ? 1 : v < 155 ? 2 : v < 195 ? 3 : 235 ? 4 : 5;
}
static int conv(int r, int g, int b) {
    return 16 + conv1(r) * 36 + conv1(g) * 6 + conv1(b);
}

static void XTERM_UpdateRects(_THIS, int numrects, SDL_Rect *rects)
{
    int x, y;
    int bytes_per_pixel = this->hidden->bpp / 8;
    char buf[7372800];
    //werase(this->hidden->win);
    //wmove(this->hidden->win, 0, 0);
    char* op = buf;
    int cnt = 0;
    strcpy(buf, "\x1b[H");
    op += 3;
    //printf("\x1b[H");
    for (y = 0; y < this->hidden->cur_h; y++) {
        for (x = 0; x < this->hidden->cur_w; x++) {
            Uint32 p = getpixel(this->hidden->buffer,
                                bytes_per_pixel,
                                this->hidden->pitch,
                                x * this->hidden->w / this->hidden->cur_w,
                                y * this->hidden->h / this->hidden->cur_h);
            if (!this->hidden->set_colored) {
                Uint8 r, g, b;
                SDL_GetRGB(p, this->hidden->fmt, &r, &g, &b);
                p = conv(r, g, b);
            }
            //printf("\x1b[48;5;%dm ", conv(r, g, b));
            if (this->hidden->text) {
                op += sprintf(op, "\x1b[48;5;%dm%c", p,
                              this->hidden->text[cnt % this->hidden->text_len]);
            }
            else {
                op += sprintf(op, "\x1b[48;5;%dm ", p);
            }

            //sprintf(buf, "\x1b[48;5;%dm ", conv(r, g, b));
            //wbkgd(this->hidden->win, COLOR_PAIR(conv(r, g, b)));
            //wattron(this->hidden->win, COLOR_PAIR(conv(r, g, b)));
            //waddstr(this->hidden->win, buf);
            //waddch(this->hidden->win, 'x');
            cnt++;
            if (x == this->hidden->cur_w - 1 &&
                y == this->hidden->cur_h - 1) break;
        }
    }
    strcpy(op, "\x1b[0m");
    printf(buf);
}

int XTERM_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors)
{
    int i;
    char buf[100*256];
    char* op = buf;
    for (i = firstcolor; i < firstcolor + ncolors; i++) {
        if (i >= 256) break;
        op += sprintf(op, "\x1b]4;%d;rgb:%2.2x/%2.2x/%2.2x\x1b\\",
                      i, colors[i].r, colors[i].g, colors[i].b);
        /*
        printf("\x1b[48;5;%dmTEST: %d %d %d %d\n", i, i,
               colors[i].r, colors[i].g, colors[i].b);
        */
    }
    this->hidden->set_colored = 1;
    printf(buf);
    return(1);
}

/* Note:  If we are terminated, this could be called in the middle of
   another SDL video routine -- notably UpdateRects.
*/
void XTERM_VideoQuit(_THIS)
{
    endwin();
    // free
}
